
#################################################*
##### ***  IMPORTS  *** #########################*
#################################################*

# *** Python modules *** #
import os

# *** Internal modules *** #
from .database import Database
from .files_manager import FileManager
from .video import Video
from .playlist import Playlist
from .__errors__ import *

#################################################*
##### ***  IMPORTS  *** #########################*
#################################################*

### *** VIDEO SERVICE CLASS *** ###
class VideoService:

    """
    VideoService
    ============
    Main class of Video Service library.

    This class will manage all the backend of the service, and you will just
    need to call it's methods when required by the server.
    """

    ## *** CLASS CONSTRUCTOR *** ##
    def __init__(self,
        DATABASE: str = ".\\Database",
        VIDEOS: str = ".\\Database\\Videos",
        THUMBNAILS: str = ".\\Database\\THUMBNAILS",
        UPLOADS: str = ".\\Uploads"
    ) -> None:

        """
        Class Constructor
        =================

        Parameters
        ----------
        DATABASE : str, optional
            Path to the database file.

        VIDEOS : str, optional
            Path to the videos folder.

        THUMBNAILS : str, optional
            Path to the thumbnails folder.

        UPLOADS : str, optional
            Path to the uploads folder.

        Raises
        ------
        TypeError
            If the paths are not strings.

        NotFoundError
            If the paths do not exist.
        """

        # Check if arguments are valid
        args_handling(
            init=True,
            DATABASE=DATABASE,
            VIDEOS=VIDEOS,
            THUMBNAILS=THUMBNAILS,
            UPLOADS=UPLOADS
        )

        # Set arguments as class attributes
        self.DB_PATH = DATABASE
        self.VIDEOS_PATH = VIDEOS
        self.THUMBNAILS_PATH = THUMBNAILS
        self.UPLOADS_PATH = UPLOADS

        # Create instances of Database and FileManager
        self.database = Database(
            DATABASE=DATABASE,
            VIDEOS=VIDEOS,
            THUMBNAILS=THUMBNAILS
        )
        self.file_manager = FileManager(
            VIDEOS=VIDEOS,
            THUMBNAILS=THUMBNAILS,
            UPLOADS=UPLOADS
        )

    ## *** CLASS METHODS *** ##
    # *** UPLOAD VIDEO TO SERVER AND DATABASE *** #
    def upload(self,
        TITLE: str,
        VIDEO_FILENAME: str,
        OWNER: str,
        VISIBILITY: str,
        THUMBNAIL_FILENAME: str = None,
        DESCRIPTION: str = None,
        TAGS: list[str] = None
    ) -> dict:

        """
        Upload
        ======
        Function to upload a video to the server and database.

        Parameters
        ----------
        TITLE : str
            Title of the video.

        VIDEO_FILENAME : str
            Filename of the video.

        OWNER : str
            Owner of the video.

        VISIBILITY : str
            Visibility of the video.

        THUMBNAIL_FILENAME : str, optional
            Filename of the thumbnail.

        DESCRIPTION : str, optional
            Description of the video.

        TAGS : list[str], optional
            Tags of the video.

        Returns
        -------
        dict
            A dictionary with the result of the operation.
        """

        try:
            # Set optional arguments as empty strings or lists
            THUMBNAIL_FILENAME = THUMBNAIL_FILENAME or ""
            DESCRIPTION = DESCRIPTION or ""
            TAGS = TAGS or [""]

            # Check if arguments are valid
            args_handling(
                init=False,
                TITLE=TITLE,
                VIDEO_FILENAME=VIDEO_FILENAME,
                OWNER=OWNER,
                VISIBILITY=VISIBILITY,
                THUMBNAIL_FILENAME=THUMBNAIL_FILENAME,
                DESCRIPTION=DESCRIPTION,
                TAGS=TAGS
            )

            # Check if files exist
            if not os.path.isfile(os.path.join(self.UPLOADS_PATH, VIDEO_FILENAME)):
                raise FileNotFoundError(f"ERROR [VideoService]: The file '{VIDEO_FILENAME}' does not exist")
            if THUMBNAIL_FILENAME and not os.path.isfile(os.path.join(self.UPLOADS_PATH, THUMBNAIL_FILENAME)):
                raise FileNotFoundError(f"ERROR [VideoService]: The file '{THUMBNAIL_FILENAME}' does not exist")

        except Exception as error:
            # Print error and return message
            error = error_parser(error)
            print(error)
            return {
                "message": error,
                "status": 500
            }

        # Upload the files to the server and get the result
        video_ = self.file_manager.upload_file(
            TITLE=TITLE,
            VIDEO_FILENAME=VIDEO_FILENAME,
            OWNER=OWNER,
            VISIBILITY=VISIBILITY,
            THUMBNAIL_FILENAME=THUMBNAIL_FILENAME,
            DESCRIPTION=DESCRIPTION,
            TAGS=TAGS
        )

        # Check if 'video_' is a dictionary
        if isinstance(video_, dict):
            return video_

        # Return the result of 'add_video' method
        return self.database.add_video(video_=video_)

    # *** CREATE PLAYLIST *** #
    def create_playlist(self,
        TITLE: str,
        OWNER: str,
        VISIBILITY: str,
        DESCRIPTION: str = None,
        TAGS: list[str] = None,
    ) -> dict:

        """
        Create Playlist
        ===============
        Function to create a playlist.

        Parameters
        ----------
        TITLE : str
            Title of the playlist.

        OWNER : str
            Owner of the playlist.

        VISIBILITY : str
            Visibility of the playlist.

        DESCRIPTION : str, optional
            Description of the playlist.

        TAGS : list[str], optional
            Tags of the playlist.

        Returns
        -------
        dict
            A dictionary with the result of the operation.
        """

        try:
            # Set optional arguments as empty strings or lists
            DESCRIPTION = DESCRIPTION or ""
            TAGS = TAGS or [""]

            # Check if arguments are valid
            args_handling(
                init=False,
                TITLE=TITLE,
                OWNER=OWNER,
                VISIBILITY=VISIBILITY,
                DESCRIPTION=DESCRIPTION,
                TAGS=TAGS
            )

        except Exception as error:
            # Print error and return message
            error = error_parser(error)
            print(error)
            return {
                "message": error,
                "status": 500
            }

        # Return the result of 'add_playlist' method
        return self.database.add_playlist(
            # Pass the result of 'create_playlist' method as 'playlist_'
            playlist_ = Playlist(
                TITLE=TITLE,
                OWNER=OWNER,
                VISIBILITY=VISIBILITY,
                DESCRIPTION=DESCRIPTION,
                TAGS=TAGS
            )
        )

    # *** SAVE VIDEOS *** #
    def save_videos(self) -> dict:

        """
        Save Videos
        ===========
        Function to save videos to the database.

        Returns
        -------
        dict
            A dictionary with the result of the operation.
        """

        # Return the result of 'save_videos' method
        return self.database.save_videos()

    # *** SAVE PLAYLISTS *** #
    def save_playlists(self) -> dict:

        """
        Save Playlists
        ==============
        Function to save playlists to the database.

        Returns
        -------
        dict
            A dictionary with the result of the operation.
        """

        # Return the result of 'save_playlists' method
        return self.database.save_playlists()

    # *** DELETE VIDEO *** #
    def delete_video(self, video_: Video) -> dict:

        """
        Delete Video
        ============
        Function to delete a video.

        Parameters
        ----------
        video_ : Video
            Video to delete.

        Returns
        -------
        dict
            A dictionary with the result of the operation.
        """

        # Return the result of 'delete_video' method
        return self.database.delete_video(video_=video_)

    # *** DELETE PLAYLIST *** #
    def delete_playlist(self, playlist_: Playlist) -> dict:

        """
        Delete Playlist
        ===============
        Function to delete a playlist.

        Parameters
        ----------
        playlist_ : Playlist
            Playlist to delete.

        Returns
        -------
        dict
            A dictionary with the result of the operation.
        """

        # Return the result of 'delete_playlist' method
        return self.database.delete_playlist(playlist_=playlist_)

    # *** REMOVE VIDEO FROM PLAYLIST *** #
    def remove_video_from_playlist(self, video_: Video, playlist_: Playlist) -> dict:

        """
        Remove Video From Playlist
        ==========================
        Function to remove a video from a playlist.

        Parameters
        ----------
        video_ : Video
            Video to remove.

        playlist_ : Playlist
            Playlist to remove the video from.

        Returns
        -------
        dict
            A dictionary with the result of the operation.
        """

        # Return the result of 'remove_video_from_playlist' method
        return self.database.remove_video_from_playlist(video_=video_, playlist_=playlist_)

    # *** ADD VIDEO TO PLAYLIST *** #
    def add_video_to_playlist(self, video_: Video, playlist_: Playlist) -> dict:

        """
        Add Video To Playlist
        =====================
        Function to add a video to a playlist.

        Parameters
        ----------
        video_ : Video
            Video to add.

        playlist_ : Playlist
            Playlist to add the video to.

        Returns
        -------
        dict
            A dictionary with the result of the operation.
        """

        # Return the result of 'add_video_to_playlist' method
        return self.database.add_video_to_playlist(video_=video_, playlist_=playlist_)

    # *** UPDATE LIKES *** #
    def update_likes(self, video_: Video, likes: int) -> dict:

        """
        Update Likes
        ============
        Function to update the likes of a video.

        Parameters
        ----------
        video_ : Video
            Video to update.

        likes : int
            Number of likes to update.

        Returns
        -------
        dict
            A dictionary with the result of the operation.
        """

        try:
            # Get the index of the video
            index: int = self.database.get_index(video_=video_)
            # Update the likes of the video
            self.database.videos[index].update_likes(likes=likes)
            # Set message
            message = {
                "video": video_,
                "message": "Likes updated successfully",
                "status": 200
            }

        except Exception as error:
            # Parse error and set message
            error = error_parser(error)
            print(error)
            message = {
                "video": video_.video,
                "message": error,
                "status": 500
            }

        return message

    # *** UPDATE VIEWS *** #
    def update_views(self, video_: Video, views: int) -> dict:

        """
        Update Views
        ============
        Function to update the views of a video.

        Parameters
        ----------
        video_ : Video
            Video to update.

        views : int
            Number of views to update.

        Returns
        -------
        dict
            A dictionary with the result of the operation.
        """

        try:
            # Get the index of the video
            index: int = self.database.get_index(video_=video_)
            # Update the views of the video
            self.database.videos[index].update_views(views=views)
            # Set message
            message = {
                "video": video_,
                "message": "Views updated successfully",
                "status": 200
            }

        except Exception as error:
            # Parse error and set message
            error = error_parser(error)
            print(error)
            message = {
                "video": video_.video,
                "message": error,
                "status": 500
            }

        return message




#################################################*
##### ***  IMPORTS  *** #########################*
#################################################*

# *** Python modules *** #
from .video import Video
from .__errors__ import *

#################################################*
##### ***  CODE  ***    #########################*
#################################################*

### *** PLAYLIST CLASS *** ###
class Playlist:

    """
    Playlist
    ========
    A class to represent a playlist object.
    """

    ## *** CLASS CONSTRUCTOR *** ##
    def __init__(self,
        TITLE: str,
        OWNER: str,
        VISIBILITY: str,
        DESCRIPTION: str = None,
        TAGS: list[str] = None,
        VIDEOS: list[dict] = None
    ) -> None:

        """
        Class Constructor
        =================

        Parameters
        ----------
        TITLE : str
            The title of the playlist.

        OWNER : str
            The owner of the playlist.

        VISIBILITY : str
            The visibility of the playlist.

        DESCRIPTION : str, optional
            The description of the playlist.

        TAGS : list[str], optional
            The tags of the playlist.

        VIDEOS : list[dict], optional
            The videos of the playlist.
        """

        # Set arguments as class attributes
        self.TITLE = TITLE
        self.OWNER = OWNER
        self.VISIBILITY = VISIBILITY
        # Set optional arguments as class attributes
        self.DESCRIPTION = DESCRIPTION or ""
        self.TAGS = TAGS or [""]
        self.VIDEOS = VIDEOS or []


    ## *** CLASS PROPERTIES *** ##
    # *** PLAYLIST JSON *** #
    @property
    def playlist(self) -> dict:

        """
        Playlist JSON
        =============
        Get a JSON object representing the playlist.

        Returns
        -------
        dict
            A JSON object representing the playlist.
        """

        # Create a playlist JSON object
        playlist_json = {
            "TITLE": self.TITLE,
            "OWNER": self.OWNER,
            "VISIBILITY": self.VISIBILITY,
            "DESCRIPTION": self.DESCRIPTION,
            "TAGS": self.TAGS,
            "VIDEOS": self.VIDEOS
        }

        return playlist_json

    ## *** CLASS METHODS *** ##
    # *** STRING REPRESENTATION *** #
    def __str__(self) -> str:
        return str(self.playlist)

    # *** ADD VIDEO *** #
    def add_video(self, video_: Video) -> None:

        """
        Add Video
        =========
        Add a video to the playlist.

        Parameters
        ----------
        video_ : Video
            The video to add to the playlist.
        """

        try:
            # Check if the video already exists in the playlist
            if self.__video_exists(video_):
                raise AlreadyExistsError("ERROR [Playlist]: The video already exists in the playlist")
            else:
                # Add the video to the playlist
                self.VIDEOS.append(video_.video)

        except Exception as error:
            # Parse error and print message
            print(error_parser(error))

    # *** REMOVE VIDEO *** #
    def remove_video(self, video_: Video) -> None:

        """
        Remove Video
        ============
        Remove a video from the playlist.

        Parameters
        ----------
        video_ : Video
            The video to remove from the playlist.
        """

        try:
            # Check if the video exists in the playlist
            if self.__video_exists(video_):
                # Remove the video from the playlist
                self.VIDEOS.remove(video_.video)
            else:
                raise NotFoundError("ERROR [Playlist]: The video does not exist in the playlist")

        except Exception as error:
            # Parse error and print message
            print(error_parser(error))

    ## *** PRIVATE METHODS *** ##
    # *** VIDEO EXISTS *** #
    def __video_exists(self, video_: Video) -> bool:
        # Get the videos from the playlist
        videos = [video for video in self.VIDEOS]
        # Check if the video exists in the playlist
        return video_.video in videos




#################################################*
##### ***  IMPORTS  *** #########################*
#################################################*

# *** External modules *** #
import cv2

# *** Python modules *** #
import shutil
import os
from datetime import datetime

# *** Internal modules *** #
from .video import Video
from .__errors__ import error_parser

#################################################*
##### ***  CODE  ***    #########################*
#################################################*

### *** FILE MANAGER CLASS *** ###
class FileManager:

    """
    FileManager
    ===========
    Manages video and thumbnail files on a server and database.
    """

    ## *** CLASS CONSTRUCTOR *** ##
    def __init__(self,
        UPLOADS: str = ".\\Uploads",
        VIDEOS: str = ".\\Database\\Videos",
        THUMBNAILS: str = ".\\Database\\THUMBNAILS"
    ) -> None:

        """
        Class Constructor
        =================

        Parameters
        ----------
        UPLOADS : str, optional
            The path to the directory where the uploaded files are stored.

        VIDEOS : str, optional
            The path to the directory where the video files are stored.

        THUMBNAILS : str, optional
            The path to the directory where the thumbnail files are stored.
        """

        # Set arguments as class attributes
        self.UPLOADS = UPLOADS
        self.VIDEOS = VIDEOS
        self.THUMBNAILS = THUMBNAILS

    ## *** CLASS METHODS *** ##
    # *** UPLOAD FILE TO SERVER AND DATABASE *** #
    def upload_file(self, **kwargs) -> Video | dict:

        """
        Upload File
        ===========
        Uploads a video and thumbnail files to the server and database.

        Parameters
        ----------
        VIDEO_FILENAME : str
            The name of the video file.

        THUMBNAIL_FILENAME : str, optional
            The name of the thumbnail file.

        TITLE : str
            The title of the video.

        OWNER : str
            The username of the video owner.

        VISIBILITY : str
            The visibility of the video.

        DESCRIPTION : str, optional
            The description of the video.

        TAGS : list, optional
            The tags of the video.

        Returns
        -------
        Video
            The video object if the files are uploaded successfully.

        dict
            A dictionary with an error message if the files are not uploaded.
        """

        # Get the video and thumbnail filenames
        video_filename = kwargs.get("VIDEO_FILENAME") or None
        thumbnail_filename = kwargs.get("THUMBNAIL_FILENAME") or None

        # Check if the video filename is provided
        if not video_filename:
            # Return an error message if the video filename is not provided
            return {
                "message": "ERROR [FileManager]: 'VIDEO_FILENAME' is required",
                "status": 400
            }

        # Get the next index for file naming
        index = self.__get_index()

        # Rename the files
        video_filename = self.__rename_file(
            file = video_filename,
            file_type = "video",
            index = index
        )

        # Check if the thumbnail filename is provided
        if not thumbnail_filename:
            # Create a thumbnail if the thumbnail filename is not provided
            thumbnail_filename = self.__create_thumbnail(
                file = video_filename,
                index = index
            )
        else:
            # Rename the thumbnail file if it is provided
            thumbnail_filename = self.__rename_file(
                file = thumbnail_filename,
                file_type = "thumbnail",
                index = index
            )

        # Upload the files to the database
        self.__upload_to_database([video_filename, thumbnail_filename])

        # Create a video object parameters
        video = {
            "TITLE": kwargs.get("TITLE"),
            "VIDEO_FILENAME": video_filename,
            "VIDEO_FILETYPE": video_filename.split('.')[1],
            "THUMBNAIL_FILENAME": thumbnail_filename,
            "THUMBNAIL_FILETYPE": thumbnail_filename.split('.')[1],
            "UPLOAD_DATE": datetime.now().strftime("%d/%m/%Y %H:%M:%S"),
            "OWNER": kwargs.get("OWNER"),
            "VISIBILITY": kwargs.get("VISIBILITY"),
            "LENGTH": self.__get_length(video_filename),
            "DESCRIPTION": kwargs.get("DESCRIPTION"),
            "TAGS": kwargs.get("TAGS")
        }

        # Return the video object
        return Video(**video)

    # *** DELETE FILE FROM SERVER AND DATABASE *** #
    @staticmethod
    def delete_file(file_name: str, directory: str) -> None:

        """
        Delete File
        ===========
        Deletes a file from the server and database.

        Parameters
        ----------
        file_name : str
            The name of the file to delete.

        directory : str
            The directory where the file is stored.
        """

        try:
            # Get the file path
            file = os.path.join(directory, file_name)
            # Remove the file
            os.remove(file)

        except Exception as error:
            # Print the error message
            print(error_parser(error))

    ## *** PRIVATE METHODS *** ##
    # *** Upload files to the database *** #
    def __upload_to_database(self, files: list) -> None:

        try:
            # Iterate over the files
            for file in files:
                # Set the source and destiny paths
                source_dist = os.path.join(self.UPLOADS, file)
                # Set the destiny path
                destiny_dist = os.path.join(
                    # Check if the file is a video or a thumbnail
                    self.VIDEOS if "video" in file else self.THUMBNAILS,
                    file
                )
                # Move the file to the destiny
                shutil.move(source_dist, destiny_dist)

        except Exception as error:
            # Print the error message
            print(error_parser(error))

    # *** Get the next index for file naming *** #
    def __get_index(self) -> int:
        # Set the index to 0
        index: int = 0

        try:
            # Get the files in the videos directory
            files = os.listdir(self.VIDEOS)
            # Get the last file
            file = files[-1]

            # Get the index from the file name
            index = int(file.split("_")[1].split(".")[0])
            # Check if the index is valid
            if index < 0:
                raise ValueError("ERROR [FileManager]: Invalid index")

        except Exception as error:
            # Print the error message
            print(error_parser(error))

        return index

    # *** Rename the file *** #
    def __rename_file(self, file_name: str, file_type: str, index: int) -> str:
        # Set the new name to None
        new_name: str = ""

        try:
            # Set the paths
            old_dist = os.path.join(self.UPLOADS, file_name) # Set the old path
            new_name = f"{file_type}_{index}.{file_name.split('.')[1]}" # Set the new name
            new_dist = os.path.join(self.UPLOADS, new_name) # Set the new path

            # Rename the file
            os.rename(old_dist, new_dist)

        except Exception as error:
            # Print the error message
            print(error_parser(error))

        return new_name

    # *** Create a thumbnail *** #
    def __create_thumbnail(self, file_name: str, index: int) -> str:
        # Set the new name to None
        new_name: str = ""

        try:
            # Open the video
            video = os.path.join(self.UPLOADS, file_name)
            videoObj = cv2.VideoCapture(video) # Create a video object

            # Check if the video is opened
            if not videoObj.isOpened():
                raise FileNotFoundError("ERROR [FileManager]: Video not found")

            # Get the frame located in the end of the first 10% of the video
            length = videoObj.get(cv2.CAP_PROP_FRAME_COUNT)
            if length == 0:
                raise ValueError("ERROR [FileManager]: Video has no frames")
            frame = length // 10 + 1
            videoObj.set(cv2.CAP_PROP_POS_FRAMES, frame)
            success, image = videoObj.read()

            # Check if the frame was found
            if not success or image is None:
                raise RuntimeError("ERROR [FileManager]: Frame not found")

            # Create the thumbnail
            new_name = f"miniature_{index}.jpg" # Set the new name
            thumbnail_path = os.path.join(self.THUMBNAILS, new_name) # Set the thumbnail path
            cv2.imwrite(thumbnail_path, image) # Save the thumbnail

        except Exception as error:
            # Print the error message
            print(error_parser(error))

        finally:
            # Release the video object
            if 'videoObj' in locals() and videoObj.isOpened():
                videoObj.release()

        return new_name

    # *** Get the video length *** #
    def __get_length(self, file_name: str) -> int:
        # Set the duration to 0
        duration: int = 0

        try:
            # Open the video
            video = os.path.join(self.VIDEOS, file_name)
            videoObj = cv2.VideoCapture(video) # Create a video object

            # Check if the video is opened
            if not videoObj.isOpened():
                raise FileNotFoundError("ERROR [FileManager]: Video not found")

            # Calculate the video duration in seconds using frames and fps
            length = videoObj.get(cv2.CAP_PROP_FRAME_COUNT) # Get the number of frames
            fps = videoObj.get(cv2.CAP_PROP_FPS) # Get the frames per second
            duration = length // fps # Calculate the duration

        except Exception as error:
            # Print the error message
            print(error_parser(error))

        finally:
            # Release the video object
            if 'videoObj' in locals() and videoObj.isOpened():
                videoObj.release()

        return duration




#################################################*
##### ***  IMPORTS  *** #########################*
#################################################*

# *** Python modules *** #
import os

#################################################*
##### ***  CODE  ***    #########################*
#################################################*

### *** ERROR CLASSES *** ###
## *** NOT FOUND ERROR *** ##
class NotFoundError(Exception):

    """
    NotFoundError
    =============
    Exception raised when a resource is not found.

    Parameters
    ----------
    message : str
        Error message.
    """

    def __init__(self, message: str):
        self.message = message
        super().__init__(self.message)

## *** ALREADY EXISTS ERROR *** ##
class AlreadyExistsError(Exception):

    """
    AlreadyExistsError
    ==================
    Exception raised when a resource already exists.

    Parameters
    ----------
    message : str
        Error message.
    """

    def __init__(self, message: str):
        self.message = message
        super().__init__(self.message)

### *** FUNCTIONS *** ###
## *** ERROR PARSER *** ##
def error_parser(error: str) -> str:

    """
    Error Parser
    ============
    Function to parse an error message.

    Parameters
    ----------
    error : str
        Error message.

    Returns
    -------
    error_str : str
        Parsed error message.
    """

    # Convert the error to a string
    error = repr(error)

    # Check if error uses ' or "
    if "('" in error:
        error_str = error.replace("('", ": ").replace("')", "") # Replace ' with :
    elif "(\"" in error:
        error_str = error.replace("(\"", ": ").replace("\")", "") # Replace " with :
    else:
        error_str = error # Return the error

    return error_str

## *** ARGS HANDLING *** ##
def args_handling(init: bool, **kwargs):

    """
    Args Handling
    =============
    Function to handle the arguments.

    Parameters
    ----------
    init : bool
        If the function is called from the __init__ method.

    **kwargs : dict
        Keyword arguments.

    Raises
    ------
    TypeError
        If all values are not of type `str`.

        If `TAGS` is not of type `list`.

        If all values in `TAGS` are not of type `str`.

    NotFoundError
        If the folder does not exist.
    """

    # Values without TAGS
    values = [value[1] for value in kwargs.items() if value[0] != "TAGS"]

    # Check if all values are strings
    if not all(isinstance(value, str) for value in values):
        raise TypeError("ERROR [VideoService]: All values must be of type 'str'")

    if init:
        if not all(os.path.exists(value) for value in values):
            raise NotFoundError("ERROR [VideoService]: The folder does not exist")

    # Check if TAGS is in kwargs
    if kwargs.get("TAGS"):

        # Check if TAGS is a list
        if not isinstance(kwargs.get("TAGS"), list):
            raise TypeError("ERROR [VideoService]: 'TAGS' must be of type 'list'")

        # Check if all values in TAGS are strings
        if not all(isinstance(tag, str) for tag in kwargs.get("TAGS")):
            raise TypeError("ERROR [VideoService]: All values in 'TAGS' must be of type 'str'")




#################################################*
##### ***  CODE  ***    #########################*
#################################################*

### *** VIDEO CLASS *** ###
class Video:

    """
    Video
    =====
    A class to represent a video.
    """

    ## *** CLASS CONSTRUCTOR *** ##
    def __init__(self,
        TITLE: str,
        VIDEO_FILENAME: str,
        VIDEO_FILETYPE: str,
        THUMBNAIL_FILENAME: str,
        THUMBNAIL_FILETYPE: str,
        UPLOAD_DATE: str,
        OWNER: str,
        VISIBILITY: str,
        LENGTH: int,
        DESCRIPTION: str = None,
        TAGS: list[str] = None,
        LIKES: int = None,
        VIEWS: int = None
    ) -> None:

        """
        Class Constructor
        =================

        Parameters
        ----------
        TITLE : str
            The title of the video.

        VIDEO_FILENAME : str
            The filename of the video.

        VIDEO_FILETYPE : str
            The filetype of the video.

        THUMBNAIL_FILENAME : str
            The filename of the thumbnail.

        THUMBNAIL_FILETYPE : str
            The filetype of the thumbnail.

        UPLOAD_DATE : str
            The upload date of the video.

        OWNER : str
            The owner of the video.

        VISIBILITY : str
            The visibility of the video.

        LENGTH : int
            The length of the video.

        DESCRIPTION : str, optional
            The description of the video.

        TAGS : list[str], optional
            The tags of the video.

        LIKES : int, optional
            The likes of the video.

        VIEWS : int, optional
            The views of the video.
        """

        # Set arguments as class attributes
        self.TITLE = TITLE
        self.VIDEO_FILENAME = VIDEO_FILENAME
        self.VIDEO_FILETYPE = VIDEO_FILETYPE
        self.THUMBNAIL_FILENAME = THUMBNAIL_FILENAME
        self.THUMBNAIL_FILETYPE = THUMBNAIL_FILETYPE
        self.UPLOAD_DATE = UPLOAD_DATE
        self.LENGTH = LENGTH
        self.OWNER = OWNER
        self.VISIBILITY = VISIBILITY
        # Set optional arguments as class attributes
        self.DESCRIPTION = DESCRIPTION or ""
        self.TAGS = TAGS or [""]
        self.LIKES = LIKES or 0
        self.VIEWS = VIEWS or 0

    ## *** CLASS PROPERTIES *** ##
    # *** VIDEO JSON *** #
    @property
    def video(self) -> dict:

        """
        Video JSON
        ==========
        Get a JSON object representing the video.

        Returns
        -------
        dict
            A JSON object representing the video.
        """

        # Create a video JSON object
        video_json = {
            "TITLE": self.TITLE,
            "VIDEO_FILENAME": self.VIDEO_FILENAME,
            "VIDEO_FILETYPE": self.VIDEO_FILETYPE,
            "THUMBNAIL_FILENAME": self.THUMBNAIL_FILENAME,
            "THUMBNAIL_FILETYPE": self.THUMBNAIL_FILETYPE,
            "UPLOAD_DATE": self.UPLOAD_DATE,
            "LENGTH": self.LENGTH,
            "DESCRIPTION": self.DESCRIPTION,
            "TAGS": self.TAGS,
            "LIKES": self.LIKES
        }

        return video_json

    ## *** CLASS METHODS *** ##
    # *** STRING REPRESENTATION *** #
    def __str__(self) -> str:
        return str(self.video)

    # *** UPDATE LIKES *** #
    def update_likes(self, likes: int) -> None:

        """
        Update Likes
        ============
        Update the likes of the video.

        Parameters
        ----------
        likes : int
            The number of likes to add to the video.
        """

        self.LIKES += likes

    # *** UPDATE VIEWS *** #
    def update_views(self, views: int) -> None:

        """
        Update Views
        ============
        Update the views of the video.

        Parameters
        ----------
        views : int
            The number of views to add to the video.
        """

        self.VIEWS += views




#################################################*
##### ***  IMPORTS  *** #########################*
#################################################*

# *** Python modules *** #
import json
import os

# *** Internal modules *** #
from .video import Video
from .playlist import Playlist
from .files_manager import FileManager
from .__errors__ import *

#################################################*
##### ***  CODE  ***    #########################*
#################################################*

### *** DATABASE CLASS *** ###
class Database:

    """
    Database
    ========
    The Database class is used to manage the videos and playlists databases.
    """

    ## *** CLASS ATTRIBUTES *** ##
    DATABASES = [
        "videos.json",
        "playlists.json"
    ]

    ## *** CLASS CONSTRUCTOR *** ##
    def __init__(self,
        DATABASE: str = ".\\Database",
        VIDEOS: str = ".\\Database\\Videos",
        THUMBNAILS: str = ".\\Database\\THUMBNAILS"
    ) -> None:

        """
        Class Constructor
        =================

        Parameters
        ----------
        DATABASE : str, optional
            The path to the database directory.

        VIDEOS : str, optional
            The path to the videos directory.

        THUMBNAILS : str, optional
            The path to the thumbnails directory.
        """

        # Set arguments as class attributes
        self.DATABASE = DATABASE
        self.VIDEOS = VIDEOS
        self.THUMBNAILS = THUMBNAILS

        # Check if directories exist
        for database in self.DATABASES:
            if not os.path.exists(f"{self.DATABASE}\\{database}"):
                # Create directories if they don't exist
                with open(f"{self.DATABASE}\\{database}", "w") as file:
                    json.dump([], file)

        # Create directories if they don't exist
        self.load_videos()
        self.load_playlists()

    ## *** CLASS METHODS *** ##
    # *** LOAD VIDEOS *** #
    def load_videos(self) -> None:

        """
        Load Videos
        ===========
        The load_videos method is used to load the videos from the database.
        """

        # Load videos from the database
        with open(f"{self.DATABASE}\\{self.DATABASES[0]}", "r") as file:
            data = json.load(file)
            self.videos = [Video(**video) for video in data]

    # *** LOAD PLAYLISTS *** #
    def load_playlists(self) -> None:

        """
        Load Playlists
        ==============
        The load_playlists method is used to load the playlists from the database.
        """

        # Load playlists from the database
        with open(f"{self.DATABASE}\\{self.DATABASES[1]}", "r") as file:
            data = json.load(file)
            self.playlists = [Playlist(**playlist) for playlist in data]

    # *** SAVE VIDEOS *** #
    def save_videos(self) -> dict:

        """
        Save Videos
        ===========
        The save_videos method is used to save the videos to the database.

        Returns
        -------
        dict
            A dictionary containing a message
        """

        try:
            # Create a list of video dictionaries
            data = [video.video for video in self.videos]
            # Save the videos to the database
            with open(f"{self.DATABASE}\\{self.DATABASES[0]}", "w") as file:
                json.dump(data, file)
            # Set message
            message = {
                "message": "Videos saved successfully",
                "status": 200
            }

        except Exception as error:
            # Print error and return message
            error = error_parser(error)
            print(error)
            message = {
                "message": error,
                "status": 500
            }

        return message

    # *** SAVE PLAYLISTS *** #
    def save_playlists(self) -> dict:

        """
        Save Playlists
        ==============
        The save_playlists method is used to save the playlists to the database.

        Returns
        -------
        dict
            A dictionary containing a message.
        """

        try:
            # Create a list of playlist dictionaries
            data = [playlist.playlist for playlist in self.playlists]
            with open(f"{self.DATABASE}\\{self.DATABASES[1]}", "w") as file:
                json.dump(data, file)
            # Set message
            message = {
                "message": "Playlists saved successfully",
                "status": 200
            }

        except Exception as error:
            # Print error and return message
            error = error_parser(error)
            print(error)
            message = {
                "message": error,
                "status": 500
            }

        return message

    # *** ADD VIDEOS *** #
    def add_video(self, video_: Video) -> dict:

        """
        Add Video
        =========
        The add_video method is used to add a video to the database.

        Parameters
        ----------
        video_ : Video
            The video object to be added to the database.

        Returns
        -------
        dict
            A dictionary containing the video object and a message.
        """

        try:
            # Create a list of video dictionaries
            video_dicts = [video.video for video in self.videos]
            # Check if video already exists
            if video_.video not in video_dicts:
                # Add video to the database
                self.videos.append(video_)
                # Set message
                message = {
                    "video": video_,
                    "message": "Video added successfully",
                    "status": 200
                }
            else:
                raise AlreadyExistsError("ERROR [DataBase]: Video already exists")

        except Exception as error:
            # Parse error and set message
            error = error_parser(error)
            print(error)
            message = {
                "video": video_.video,
                "message": error,
                "status": 500
            }

        # Save the videos to the database
        self.save_videos()
        return message

    # *** ADD PLAYLISTS *** #
    def add_playlist(self, playlist_: Playlist) -> dict:

        """
        Add Playlist
        ============
        The add_playlist method is used to add a playlist to the database.

        Parameters
        ----------
        playlist_ : Playlist
            The playlist object to be added to the database.

        Returns
        -------
        dict
            A dictionary containing the playlist object and a message.
        """

        try:
            # Create a list of playlist dictionaries
            playlist_dicts = [playlist.playlist for playlist in self.playlists]
            # Check if playlist already exists
            if playlist_.playlist not in playlist_dicts:
                # Add playlist to the database
                self.playlists.append(playlist_)
                message = {
                    "playlist": playlist_,
                    "message": "Playlist added successfully",
                    "status": 200
                }
            else:
                raise AlreadyExistsError("ERROR [DataBase]: Video already exists")

        except Exception as error:
            # Parse error and set message
            error = error_parser(error)
            print(error)
            message = {
                "playlist": playlist_.playlist,
                "message": error,
                "status": 500
            }

        # Save the playlists to the database
        self.save_playlists()
        return message

    # *** DELETE VIDEOS *** #
    def delete_video(self, video_: Video) -> dict:

        """
        Delete Video
        ============
        The delete_video method is used to delete a video from the database.

        Parameters
        ----------
        video_ : Video
            The video object to be deleted from the database.

        Returns
        -------
        dict
            A dictionary containing the video object and a message.
        """

        try:
            # Create a list with the video and thumbnail filenames
            files = [
                [video_.video.get("VIDEO_FILENAME"), self.VIDEOS],
                [video_.video.get("THUMBNAIL_FILENAME"), self.THUMBNAILS]
            ]

            # Check if files exist
            for file in files:
                if file[0] not in os.listdir(file[1]):
                    raise FileNotFoundError("ERROR [DataBase]: File not found in the directory")
                FileManager.delete_file(*file)

            # Get the video index
            index = self.get_index(video_=video_)
            # Delete the video
            self.videos.pop(index)
            # Set message
            message = {
                "video": video_,
                "message": "Video deleted successfully",
                "status": 200
            }

        except Exception as error:
            # Parse error and set message
            error = error_parser(error)
            print(error)
            message = {
                "video": video_.video,
                "message": error,
                "status": 500
            }

        # Save the videos to the database
        self.save_videos()
        return message

    # *** DELETE PLAYLISTS *** #
    def delete_playlist(self, playlist_: Playlist) -> dict:

        """
        Delete Playlist
        ===============
        The delete_playlist method is used to delete a playlist from the database.

        Parameters
        ----------
        playlist_ : Playlist
            The playlist object to be deleted from the database.

        Returns
        -------
        dict
            A dictionary containing the playlist object and a message
        """

        try:
            # Get the playlist index
            index = self.get_index(playlist_=playlist_)
            # Delete the playlist
            self.playlists.pop(index)
            # Set message
            message = {
                "playlist": playlist_,
                "message": "Playlist deleted successfully",
                "status": 200
            }

        except Exception as error:
            # Parse error and set message
            error = error_parser(error)
            print(error)
            message = {
                "playlist": playlist_.playlist,
                "message": error,
                "status": 500
            }

        # Save the playlists to the database
        self.save_playlists()
        return message

    # *** ADD VIDEO TO PLAYLIST *** #
    def add_video_to_playlist(self, playlist_: Playlist, video_: Video) -> dict:

        """
        Add Video to Playlist
        =====================
        The add_video_to_playlist method is used to add a video to a playlist.

        Parameters
        ----------
        playlist_ : Playlist
            The playlist object to add the video to.

        video_ : Video
            The video object to be added to the playlist.

        Returns
        -------
        dict
            A dictionary containing the playlist object and a message.
        """

        try:
            # Get the playlist index
            index: int = self.get_index(playlist_=playlist_)
            # Add video to playlist
            self.playlists[index].add_video(video_)
            # Set message
            message = {
                "playlist": playlist_,
                "message": "Video added to playlist successfully",
                "status": 200
            }

        except Exception as error:
            # Parse error and set message
            error = error_parser(error)
            print(error)
            message = {
                "playlist": playlist_.playlist,
                "message": error,
                "status": 500
            }

        # Save the playlists to the database
        self.save_playlists()
        return message

    # *** REMOVE VIDEO FROM PLAYLIST *** #
    def remove_video_from_playlist(self, playlist_: Playlist, video_: Video) -> dict:

        """
        Remove Video from Playlist
        ==========================
        The remove_video_from_playlist method is used to remove a video from a playlist.

        Parameters
        ----------
        playlist_ : Playlist
            The playlist object to remove the video from.

        video_ : Video
            The video object to be removed from the playlist.

        Returns
        -------
        dict
            A dictionary containing the playlist object and a message.
        """

        try:
            # Get the playlist index
            index: int = self.get_index(playlist_=playlist_)
            # Remove video from playlist
            self.playlists[index].remove_video(video_)
            # Set message
            message = {
                "playlist": playlist_,
                "message": "Video removed from playlist successfully",
                "status": 200
            }

        except Exception as error:
            # Parse error and set message
            error = error_parser(error)
            print(error)
            message = {
                "playlist": playlist_.playlist,
                "message": error,
                "status": 500
            }

        # Save the playlists to the database
        self.save_playlists()
        return message

    # *** GET INDEX *** #
    def get_index(self, video_: Video, playlist_: Video) -> int:

        """
        Get Index
        =========
        The get_index method is used to get the index of a video or playlist.

        Parameters
        ----------
        video_ : Video
            The video object to get the index of.

        playlist_ : Playlist
            The playlist object to get the index of.

        Returns
        -------
        int
            The index of the video or playlist.
        """

        # Check if video or playlist is provided
        if video_:
            # Get the video index
            return self.__get_video_index(video_)
        elif playlist_:
            # Get the playlist index
            return self.__get_playlist_index(playlist_)

    ## *** PRIVATE METHODS *** ##
    # *** GET VIDEO INDEX *** #
    def __get_video_index(self, video_: Video) -> int:
        # Create a list of video filenames
        videos = [video.video for video in self.videos]
        # Check if video exists
        if video_.video in videos:
            return videos.index(video_.video)
        else:
            raise NotFoundError("ERROR [DataBase]: Video not found in the database")

    # *** GET PLAYLIST INDEX *** #
    def __get_playlist_index(self, playlist_: Playlist) -> int:
        # Create a list of playlist filenames
        playlists = [playlist.playlist for playlist in self.playlists]
        # Check if playlist exists
        if playlist_.playlist in playlists:
            return playlists.index(playlist_.playlist)
        else:
            raise NotFoundError("ERROR [DataBase]: Playlist not found in the database")
